@@ -27,18 +27,23 @@
 #define RSVG_PRIVATE_H
 
 #include "rsvg.h"
-#include "rsvg-styles.h"
 
 #include <libxml/SAX.h>
 #include <libxml/xmlmemory.h>
+#include <pango/pangoft2.h>
 
 G_BEGIN_DECLS
 
 typedef struct RsvgSaxHandler RsvgSaxHandler;
+typedef struct _RsvgPropertyBag RsvgPropertyBag;
+typedef struct _RsvgState RsvgState;
+typedef struct _RsvgDefs RsvgDefs;
+typedef struct _RsvgDefVal RsvgDefVal;
+typedef struct _RsvgFilter RsvgFilter;
 
 struct RsvgSaxHandler {
 	void (*free) (RsvgSaxHandler *self);
-	void (*start_element) (RsvgSaxHandler *self, const xmlChar *name, const xmlChar **atts);
+	void (*start_element) (RsvgSaxHandler *self, const xmlChar *name, RsvgPropertyBag *atts);
 	void (*end_element) (RsvgSaxHandler *self, const xmlChar *name);
 	void (*characters) (RsvgSaxHandler *self, const xmlChar *ch, int len);
 };
@@ -56,6 +61,8 @@ struct RsvgHandle {
 	
 	RsvgDefs *defs;
 	gboolean in_defs;
+	void *current_defs_group;
+
 	GHashTable *css_props;
 	
 	/* not a handler stack. each nested handler keeps
@@ -77,6 +84,9 @@ struct RsvgHandle {
 	GString * title;
 	GString * desc;
 
+	void * currentfilter;
+	void * currentsubfilter;
+
 	/* virtual fns */
 	gboolean (* write) (RsvgHandle    *handle,
 						const guchar  *buf,
@@ -102,6 +112,50 @@ gboolean rsvg_handle_close_impl (RsvgHandle  *handle,
 								 GError     **error);
 void rsvg_handle_free_impl (RsvgHandle *handle);
 
+typedef enum {
+	RSVG_SIZE_ZOOM,
+	RSVG_SIZE_WH,
+	RSVG_SIZE_WH_MAX,
+	RSVG_SIZE_ZOOM_MAX
+} RsvgSizeType;
+
+struct RsvgSizeCallbackData
+{
+	RsvgSizeType type;
+	double x_zoom;
+	double y_zoom;
+	gint width;
+	gint height;
+};
+
+/* private */
+GdkPixbuf *
+rsvg_pixbuf_from_file_with_size_data (const gchar * file_name,
+									  struct RsvgSizeCallbackData * data,
+									  GError ** error);
+/* private */
+GdkPixbuf *
+rsvg_pixbuf_from_stdio_file_with_size_data(FILE * f,
+										   struct RsvgSizeCallbackData * data,
+										   GError ** error);
+
+struct _RsvgPropertyBag
+{
+	GHashTable * props;
+};
+
+RsvgPropertyBag *
+rsvg_property_bag_new (const xmlChar **atts);
+
+void
+rsvg_property_bag_free (RsvgPropertyBag *bag);
+
+const char *
+rsvg_property_bag_lookup (RsvgPropertyBag *bag, const char * key);
+
+guint
+rsvg_property_bag_size (RsvgPropertyBag *bag);
+
 G_END_DECLS
 
 #endif
