@@ -25,10 +25,10 @@
 #include <string.h>
 #include <math.h>
 
-#include "rsvg-shapes.h"
-#include "rsvg-styles.h"
-#include "rsvg-css.h"
 #include "rsvg-private.h"
+#include "rsvg-styles.h"
+#include "rsvg-shapes.h"
+#include "rsvg-css.h"
 #include "rsvg-bpath-util.h"
 #include "rsvg-path.h"
 #include "rsvg-defs.h"
@@ -45,13 +45,26 @@
 /* 4/3 * (1-cos 45)/sin 45 = 4/3 * sqrt(2) - 1 */
 #define RSVG_ARC_MAGIC ((double) 0.5522847498)
 
-struct _RsvgDefsPath {
- 	RsvgDefVal super;
-	RsvgState  state;
+
+typedef struct _RsvgDefsDrawablePath RsvgDefsDrawablePath;
+typedef struct _RsvgDefsDrawableGroup RsvgDefsDrawableGroup;
+typedef struct _RsvgDefsDrawableUse RsvgDefsDrawableUse;
+
+struct _RsvgDefsDrawablePath {
+ 	RsvgDefsDrawable super;
  	char       *d;
 };
- 
-typedef struct _RsvgDefsPath RsvgDefsPath;
+
+struct _RsvgDefsDrawableGroup {
+ 	RsvgDefsDrawable super;
+ 	GPtrArray *children;
+};
+
+struct _RsvgDefsDrawableUse {
+ 	RsvgDefsDrawable super;
+ 	RsvgDefsDrawable *child;
+};
+
 
 /**
  * rsvg_close_vpath: Close a vector path.
@@ -181,7 +194,7 @@ rsvg_render_svp (RsvgHandle *ctx, const ArtSVP *svp,
 		}
 	
 	has_alpha = gdk_pixbuf_get_has_alpha (pixbuf);
-	
+
 	render = art_render_new (0, 0,
 							 gdk_pixbuf_get_width (pixbuf),
 							 gdk_pixbuf_get_height (pixbuf),
@@ -192,7 +205,7 @@ rsvg_render_svp (RsvgHandle *ctx, const ArtSVP *svp,
 							 gdk_pixbuf_get_bits_per_sample (pixbuf),
 							 has_alpha ? ART_ALPHA_SEPARATE : ART_ALPHA_NONE,
 							 NULL);
-	
+
 	art_render_svp (render, svp);
 	art_render_mask_solid (render, (opacity << 8) + opacity + (opacity >> 7));
 
@@ -206,7 +219,6 @@ rsvg_render_svp (RsvgHandle *ctx, const ArtSVP *svp,
 	state = rsvg_state_current(ctx);
 	for (i = 0; i < 6; i++)
 		gradctx.affine[i] = state->affine[i];
-
 	rsvg_render_paint_server (render, ps, &gradctx);
 	art_render_invoke (render);
 }
@@ -244,11 +256,11 @@ rsvg_render_bpath (RsvgHandle *ctx, const ArtBpath *bpath)
 	vpath = art_bez_path_to_vec (affine_bpath, 0.25);
 	art_free (affine_bpath);
 	
-	need_tmpbuf = (state->fill != NULL) && (state->stroke != NULL) &&
-		state->opacity != 0xff;
+	need_tmpbuf = ((state->fill != NULL) && (state->stroke != NULL) &&
+				   state->opacity != 0xff) || state->filter;
 	
 	if (need_tmpbuf)
-		rsvg_push_opacity_group (ctx);
+		rsvg_push_discrete_layer (ctx);
 	
 	if (state->fill != NULL)
 		{
@@ -308,9 +320,9 @@ rsvg_render_bpath (RsvgHandle *ctx, const ArtBpath *bpath)
 			rsvg_render_svp (ctx, svp, state->stroke, opacity);
 			art_svp_free (svp);
 		}
-	
+
 	if (need_tmpbuf)
-		rsvg_pop_opacity_group (ctx, state->opacity);
+		rsvg_pop_discrete_layer (ctx);	
 	
 	art_free (vpath);
 }
@@ -328,59 +340,224 @@ rsvg_render_path(RsvgHandle *ctx, const char *d)
 	rsvg_bpath_def_free (bpath_def);
 }
 
-static void 
-rsvg_defs_path_free (RsvgDefVal *self)
+void 
+rsvg_defs_drawable_draw (RsvgDefsDrawable * self, RsvgHandle *ctx)
 {
-	RsvgDefsPath *z = (RsvgDefsPath *)self;
-	rsvg_state_finalize (&z->state);
+	self->draw(self, ctx);
+}
+
+static void 
+rsvg_defs_drawable_path_free (RsvgDefVal *self)
+{
+	RsvgDefsDrawablePath *z = (RsvgDefsDrawablePath *)self;
+	rsvg_state_finalize (&z->super.state);
 	g_free (z);
 }
 
+static void 
+rsvg_defs_drawable_path_draw (RsvgDefsDrawable * self, RsvgHandle *ctx)
+{
+	RsvgState *state = rsvg_state_current (ctx);
+	RsvgDefsDrawablePath *path = (RsvgDefsDrawablePath*)self;
+
+	/* combine state definitions */
+
+	rsvg_state_clone (state, &self->state);
+	if (ctx->n_state > 1)
+		rsvg_state_reinherit(state, &ctx->state[ctx->n_state - 2]);
+	
+	if (state->opacity != 0xff || state->filter)
+		rsvg_push_discrete_layer (ctx);
+	
+	/* always want to render inside of a <use/> */
+	rsvg_render_path (ctx, path->d);
+	
+	if (state->opacity != 0xff || state->filter)
+		rsvg_pop_discrete_layer (ctx);
+}
+
+static void 
+rsvg_defs_drawable_group_free (RsvgDefVal *self)
+{
+	RsvgDefsDrawableGroup *z = (RsvgDefsDrawableGroup *)self;
+	rsvg_state_finalize (&z->super.state);
+	g_ptr_array_free(z->children, FALSE);
+	g_free (z);
+}
+
+static void 
+rsvg_defs_drawable_group_draw (RsvgDefsDrawable * self, RsvgHandle *ctx)
+{
+	RsvgState *state = rsvg_state_current (ctx);
+	RsvgDefsDrawableGroup *group = (RsvgDefsDrawableGroup*)self;
+	guint i;
+
+	/* combine state definitions */
+	rsvg_state_clone (state, &self->state);
+	if (ctx->n_state > 1)
+		rsvg_state_reinherit(state, &ctx->state[ctx->n_state - 2]);
+	
+	if (state->opacity != 0xff || state->filter)
+		rsvg_push_discrete_layer (ctx);
+
+	for (i = 0; i < group->children->len; i++)
+		{
+			/* push the state stack */
+			if (ctx->n_state == ctx->n_state_max)
+				ctx->state = g_renew (RsvgState, ctx->state, 
+									  ctx->n_state_max <<= 1);
+			if (ctx->n_state)
+				rsvg_state_inherit (&ctx->state[ctx->n_state],
+									&ctx->state[ctx->n_state - 1]);
+			else
+				rsvg_state_init (ctx->state);
+			ctx->n_state++;
+			
+			rsvg_defs_drawable_draw (g_ptr_array_index(group->children, i), ctx);
+	
+			/* pop the state stack */
+			ctx->n_state--;
+			rsvg_state_finalize (&ctx->state[ctx->n_state]);
+		}			
+
+	if (state->opacity != 0xff || state->filter)
+		rsvg_pop_discrete_layer (ctx);
+	
+}
+
+static void 
+rsvg_defs_drawable_use_free (RsvgDefVal *self)
+{
+	RsvgDefsDrawableUse *z = (RsvgDefsDrawableUse *)self;
+	rsvg_state_finalize (&z->super.state);
+	g_free (z);
+}
+
+static void 
+rsvg_defs_drawable_use_draw (RsvgDefsDrawable * self, RsvgHandle *ctx)
+{
+	RsvgState *state = rsvg_state_current (ctx);
+	RsvgDefsDrawableUse *use = (RsvgDefsDrawableUse*)self;
+
+	/* combine state definitions */
+	
+	rsvg_state_clone (state, &self->state);
+	if (ctx->n_state > 1)
+		rsvg_state_reinherit(state, &ctx->state[ctx->n_state - 2]);
+	
+	if (state->opacity != 0xff || state->filter)
+		rsvg_push_discrete_layer (ctx);
+
+
+	/* push the state stack */
+	if (ctx->n_state == ctx->n_state_max)
+		ctx->state = g_renew (RsvgState, ctx->state, 
+							  ctx->n_state_max <<= 1);
+	if (ctx->n_state)
+		rsvg_state_inherit (&ctx->state[ctx->n_state],
+							&ctx->state[ctx->n_state - 1]);
+	else
+		rsvg_state_init (ctx->state);
+	ctx->n_state++;
+	
+	rsvg_defs_drawable_draw (use->child, ctx);
+	
+	/* pop the state stack */
+	ctx->n_state--;
+	rsvg_state_finalize (&ctx->state[ctx->n_state]);
+
+	if (state->opacity != 0xff || state->filter)
+		rsvg_pop_discrete_layer (ctx);
+}			
+
+static void 
+rsvg_defs_drawable_group_pack (RsvgDefsDrawableGroup *self, RsvgDefsDrawable *child)
+{
+	RsvgDefsDrawableGroup *z = (RsvgDefsDrawableGroup *)self;
+	g_ptr_array_add(z->children, child);
+}
+
+void
+rsvg_push_def_group (RsvgHandle *ctx, const char * id)
+{
+	RsvgDefsDrawableGroup *group;
+	if (!ctx->in_defs)
+		return;	
+
+	group = g_new (RsvgDefsDrawableGroup, 1);
+	group->children = g_ptr_array_new();
+	rsvg_state_clone (&group->super.state, rsvg_state_current (ctx));
+
+	group->super.super.type = RSVG_DEF_PATH;
+	group->super.super.free = rsvg_defs_drawable_group_free;
+	group->super.draw = rsvg_defs_drawable_group_draw;
+
+	if (id != NULL)
+		rsvg_defs_set (ctx->defs, id, &group->super.super);
+
+	group->super.parent = (RsvgDefsDrawable *)ctx->current_defs_group;
+	if (group->super.parent != NULL)
+		rsvg_defs_drawable_group_pack((RsvgDefsDrawableGroup *)group->super.parent, 
+									  &group->super);
+	ctx->current_defs_group = group;
+}
+
+void
+rsvg_pop_def_group (RsvgHandle *ctx)
+{
+	RsvgDefsDrawableGroup * group;
+	if (!ctx->in_defs)
+		return;
+	group = (RsvgDefsDrawableGroup *)ctx->current_defs_group;
+	if (group == NULL)
+		return;
+	ctx->current_defs_group = group->super.parent;
+}
+
 void
 rsvg_handle_path (RsvgHandle *ctx, const char * d, const char * id)
 {
 	if (!ctx->in_defs)
 		rsvg_render_path (ctx, d);
 	else {
-		RsvgDefsPath *path;
+		RsvgDefsDrawablePath *path;
 
-		if (id == NULL) 
-			return;
-
-		path = g_new (RsvgDefsPath, 1);
+		path = g_new (RsvgDefsDrawablePath, 1);
 		path->d = g_strdup(d);
-		rsvg_state_clone (&path->state, rsvg_state_current (ctx));
-		path->super.type = RSVG_DEF_PATH;
-		path->super.free = rsvg_defs_path_free;
-		
-		rsvg_defs_set (ctx->defs, id, &path->super);
+		rsvg_state_clone (&path->super.state, rsvg_state_current (ctx));
+		path->super.super.type = RSVG_DEF_PATH;
+		path->super.super.free = rsvg_defs_drawable_path_free;
+		path->super.draw = rsvg_defs_drawable_path_draw;
+		if (id != NULL)
+			rsvg_defs_set (ctx->defs, id, &path->super.super);
+
+		path->super.parent = (RsvgDefsDrawable *)ctx->current_defs_group;
+		if (path->super.parent != NULL)
+			rsvg_defs_drawable_group_pack((RsvgDefsDrawableGroup *)path->super.parent, 
+										  &path->super);
 	}
 }
 
 void
-rsvg_start_path (RsvgHandle *ctx, const xmlChar **atts)
+rsvg_start_path (RsvgHandle *ctx, RsvgPropertyBag *atts)
 {
-	int i;
-	char *d = NULL;
-	const char * klazz = NULL, * id = NULL;
+	const char * klazz = NULL, * id = NULL, *value, *d = NULL;
 	
-	if (atts != NULL)
+	if (rsvg_property_bag_size (atts))
 		{
-			for (i = 0; atts[i] != NULL; i += 2)
-				{
-					if (!strcmp ((char *)atts[i], "d"))
-						d = (char *)atts[i + 1];
-					else if (!strcmp ((char *)atts[i], "class"))
-						klazz = (char *)atts[i + 1];
-					else if (!strcmp ((char *)atts[i], "id"))
-						id = (const char *)atts[i + 1];
-				}
+			if ((value = rsvg_property_bag_lookup (atts, "d")))
+				d = value;
+			if ((value = rsvg_property_bag_lookup (atts, "class")))
+				klazz = value;
+			if ((value = rsvg_property_bag_lookup (atts, "id")))
+				id = value;
+
+			rsvg_parse_style_attrs (ctx, rsvg_state_current (ctx), "path", klazz, id, atts);
 		}
 	
 	if (d == NULL)
 		return;
 	
-	rsvg_parse_style_attrs (ctx, rsvg_state_current (ctx), "path", klazz, id, atts);
 	rsvg_handle_path (ctx, d, id);
 }
 
@@ -410,35 +587,31 @@ rsvg_make_poly_point_list(const char * points)
 }
 
 static void
-rsvg_start_any_poly(RsvgHandle *ctx, const xmlChar **atts, gboolean is_polyline)
+rsvg_start_any_poly(RsvgHandle *ctx, RsvgPropertyBag *atts, gboolean is_polyline)
 {
 	/* the only difference between polygon and polyline is
 	   that a polyline closes the path */
 	
-	int i;
 	const char * verts = (const char *)NULL;
 	GString * g = NULL;
 	gchar ** pointlist = NULL;
-	const char * klazz = NULL, * id = NULL;
+	const char * klazz = NULL, * id = NULL, *value;
 
-	if (atts != NULL)
+	if (rsvg_property_bag_size (atts))
 		{
-			for (i = 0; atts[i] != NULL; i += 2)
-				{
-					/* support for svg < 1.0 which used verts */
-					if (!strcmp ((char *)atts[i], "verts") || !strcmp ((char *)atts[i], "points"))
-						verts = (const char *)atts[i + 1];
-					else if (!strcmp ((char *)atts[i], "class"))
-						klazz = (const char *)atts[i + 1];
-					else if (!strcmp ((char *)atts[i], "id"))
-						id = (const char *)atts[i + 1];
-				}
+			/* support for svg < 1.0 which used verts */
+			if ((value = rsvg_property_bag_lookup (atts, "verts")) || (value = rsvg_property_bag_lookup (atts, "points")))
+				verts = value;
+			if ((value = rsvg_property_bag_lookup (atts, "class")))
+				klazz = value;
+			if ((value = rsvg_property_bag_lookup (atts, "id")))
+				id = value;
+
+			rsvg_parse_style_attrs (ctx, rsvg_state_current (ctx), (is_polyline ? "polyline" : "polygon"), klazz, id, atts);
 		}
 	
 	if (!verts)
-		return;
-	
-	rsvg_parse_style_attrs (ctx, rsvg_state_current (ctx), (is_polyline ? "polyline" : "polygon"), klazz, id, atts);
+		return;	
 	
 	/* todo: make the following more memory and CPU friendly */
 	g = rsvg_make_poly_point_list (verts);
@@ -448,6 +621,7 @@ rsvg_start_any_poly(RsvgHandle *ctx, const xmlChar **atts, gboolean is_polyline)
 	/* represent as a "moveto, lineto*, close" path */  
 	if (pointlist)
 		{
+			int i;
 			GString * d = g_string_sized_new (strlen(verts));
 			g_string_append_printf (d, "M %s %s ", pointlist[0], pointlist[1] );
 			
@@ -464,51 +638,45 @@ rsvg_start_any_poly(RsvgHandle *ctx, const xmlChar **atts, gboolean is_polyline)
 }
 
 void
-rsvg_start_polygon (RsvgHandle *ctx, const xmlChar **atts)
+rsvg_start_polygon (RsvgHandle *ctx, RsvgPropertyBag *atts)
 {
 	rsvg_start_any_poly (ctx, atts, FALSE);
 }
 
 void
-rsvg_start_polyline (RsvgHandle *ctx, const xmlChar **atts)
+rsvg_start_polyline (RsvgHandle *ctx, RsvgPropertyBag *atts)
 {
 	rsvg_start_any_poly (ctx, atts, TRUE);
 }
 
 void
-rsvg_start_line (RsvgHandle *ctx, const xmlChar **atts)
+rsvg_start_line (RsvgHandle *ctx, RsvgPropertyBag *atts)
 {
-	int i;
 	double x1 = 0, y1 = 0, x2 = 0, y2 = 0;
 	GString * d = NULL;
-	const char * klazz = NULL, * id = NULL;
+	const char * klazz = NULL, * id = NULL, *value;
 	char buf [G_ASCII_DTOSTR_BUF_SIZE];
 	double font_size;
 
-	if (ctx->n_state > 0)
-		font_size = rsvg_state_current (ctx)->font_size;
-	else
-		font_size = 12.0;
+	font_size = rsvg_state_current_font_size (ctx);
 
-	if (atts != NULL)
+	if (rsvg_property_bag_size (atts))
 		{
-			for (i = 0; atts[i] != NULL; i += 2)
-				{
-					if (!strcmp ((char *)atts[i], "x1"))
-						x1 = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, font_size);
-					else if (!strcmp ((char *)atts[i], "y1"))
-						y1 = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, font_size);
-					if (!strcmp ((char *)atts[i], "x2"))
-						x2 = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, font_size);
-					else if (!strcmp ((char *)atts[i], "y2"))
-						y2 = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, font_size);
-					else if (!strcmp ((char *)atts[i], "class"))
-						klazz = (const char *)atts[i + 1];
-					else if (!strcmp ((char *)atts[i], "id"))
-						id = (const char *)atts[i + 1];
-				}      
+			if ((value = rsvg_property_bag_lookup (atts, "x1")))
+				x1 = rsvg_css_parse_normalized_length (value, ctx->dpi, (gdouble)ctx->width, font_size);
+			if ((value = rsvg_property_bag_lookup (atts, "y1")))
+				y1 = rsvg_css_parse_normalized_length (value, ctx->dpi, (gdouble)ctx->height, font_size);
+			if ((value = rsvg_property_bag_lookup (atts, "x2")))
+				x2 = rsvg_css_parse_normalized_length (value, ctx->dpi, (gdouble)ctx->width, font_size);
+			if ((value = rsvg_property_bag_lookup (atts, "y2")))
+				y2 = rsvg_css_parse_normalized_length (value, ctx->dpi, (gdouble)ctx->height, font_size);
+			if ((value = rsvg_property_bag_lookup (atts, "class")))
+				klazz = value;
+			if ((value = rsvg_property_bag_lookup (atts, "id")))
+				id = value;
+
+			rsvg_parse_style_attrs (ctx, rsvg_state_current (ctx), "line", klazz, id, atts);
 		}
-	rsvg_parse_style_attrs (ctx, rsvg_state_current (ctx), "line", klazz, id, atts);
 	
 	/* emulate a line using a path */
 	/* ("M %f %f L %f %f", x1, y1, x2, y2) */
@@ -527,46 +695,41 @@ rsvg_start_line (RsvgHandle *ctx, const xmlChar **atts)
 }
 
 void
-rsvg_start_rect (RsvgHandle *ctx, const xmlChar **atts)
+rsvg_start_rect (RsvgHandle *ctx, RsvgPropertyBag *atts)
 {
-	int i;
 	double x = 0., y = 0., w = 0, h = 0, rx = 0., ry = 0.;
 	GString * d = NULL;
-	const char * klazz = NULL, * id = NULL;
+	const char * klazz = NULL, * id = NULL, *value;
 	char buf [G_ASCII_DTOSTR_BUF_SIZE];
 	gboolean got_rx = FALSE, got_ry = FALSE;
 	double font_size;
 
-	if (ctx->n_state > 0)
-		font_size = rsvg_state_current (ctx)->font_size;
-	else
-		font_size = 12.0;
+	font_size = rsvg_state_current_font_size (ctx);
 
-	if (atts != NULL)
+	if (rsvg_property_bag_size (atts))
 		{
-			for (i = 0; atts[i] != NULL; i += 2)
-				{
-					if (!strcmp ((char *)atts[i], "x"))
-						x = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, font_size);
-					else if (!strcmp ((char *)atts[i], "y"))
-						y = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, font_size);
-					else if (!strcmp ((char *)atts[i], "width"))
-						w = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, font_size);
-					else if (!strcmp ((char *)atts[i], "height"))
-						h = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, font_size);
-					else if (!strcmp ((char *)atts[i], "rx")) {
-						rx = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, font_size);
-						got_rx = TRUE;
-					}
-					else if (!strcmp ((char *)atts[i], "ry")) {
-						ry = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, font_size);
-						got_ry = TRUE;
-					}
-					else if (!strcmp ((char *)atts[i], "class"))
-						klazz = (const char *)atts[i + 1];
-					else if (!strcmp ((char *)atts[i], "id"))
-						id = (const char *)atts[i + 1];
-				}
+			if ((value = rsvg_property_bag_lookup (atts, "x")))
+				x = rsvg_css_parse_normalized_length (value, ctx->dpi, (gdouble)ctx->width, font_size);
+			if ((value = rsvg_property_bag_lookup (atts, "y")))
+				y = rsvg_css_parse_normalized_length (value, ctx->dpi, (gdouble)ctx->height, font_size);
+			if ((value = rsvg_property_bag_lookup (atts, "width")))
+				w = rsvg_css_parse_normalized_length (value, ctx->dpi, (gdouble)ctx->width, font_size);
+			if ((value = rsvg_property_bag_lookup (atts, "height")))
+				h = rsvg_css_parse_normalized_length (value, ctx->dpi, (gdouble)ctx->height, font_size);
+			if ((value = rsvg_property_bag_lookup (atts, "rx"))) {
+				rx = rsvg_css_parse_normalized_length (value, ctx->dpi, (gdouble)ctx->width, font_size);
+				got_rx = TRUE;
+			}
+			if ((value = rsvg_property_bag_lookup (atts, "ry"))) {
+				ry = rsvg_css_parse_normalized_length (value, ctx->dpi, (gdouble)ctx->height, font_size);
+				got_ry = TRUE;
+			}
+			if ((value = rsvg_property_bag_lookup (atts, "class")))
+				klazz = value;
+			if ((value = rsvg_property_bag_lookup (atts, "id")))
+				id = value;
+
+			rsvg_parse_style_attrs (ctx, rsvg_state_current (ctx), "rect", klazz, id, atts);
 		}
 
 	if (got_rx && !got_ry)
@@ -580,9 +743,7 @@ rsvg_start_rect (RsvgHandle *ctx, const xmlChar **atts)
 	if (rx > fabs(w / 2.))
 		rx = fabs(w / 2.);
 	if (ry > fabs(h / 2.))
-		ry = fabs(h / 2.);
-	
-	rsvg_parse_style_attrs (ctx, rsvg_state_current (ctx), "rect", klazz, id, atts);
+		ry = fabs(h / 2.);   
 	
 	/* incrementing y by 1 properly draws borders. this is a HACK */
 	y += .01;
@@ -671,43 +832,36 @@ rsvg_start_rect (RsvgHandle *ctx, const xmlChar **atts)
 }
 
 void
-rsvg_start_circle (RsvgHandle *ctx, const xmlChar **atts)
+rsvg_start_circle (RsvgHandle *ctx, RsvgPropertyBag *atts)
 {
-	int i;
 	double cx = 0, cy = 0, r = 0;
 	GString * d = NULL;
-	const char * klazz = NULL, * id = NULL;
+	const char * klazz = NULL, * id = NULL, *value;
 	char buf [G_ASCII_DTOSTR_BUF_SIZE];
 	double font_size;
 	
-	if (ctx->n_state > 0)
-		font_size = rsvg_state_current (ctx)->font_size;
-	else
-		font_size = 12.0;
+	font_size = rsvg_state_current_font_size (ctx);
 
-	if (atts != NULL)
+	if (rsvg_property_bag_size (atts))
 		{
-			for (i = 0; atts[i] != NULL; i += 2)
-				{
-					if (!strcmp ((char *)atts[i], "cx"))
-						cx = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, font_size);
-					else if (!strcmp ((char *)atts[i], "cy"))
-						cy = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, font_size);
-					else if (!strcmp ((char *)atts[i], "r"))
-						r = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, 
-															  rsvg_viewport_percentage((gdouble)ctx->width, (gdouble)ctx->height), 
-															  font_size);
-					else if (!strcmp ((char *)atts[i], "class"))
-						klazz = (const char *)atts[i + 1];
-					else if (!strcmp ((char *)atts[i], "id"))
-						id = (const char *)atts[i + 1];
-				}
+			if ((value = rsvg_property_bag_lookup (atts, "cx")))
+				cx = rsvg_css_parse_normalized_length (value, ctx->dpi, (gdouble)ctx->width, font_size);
+			if ((value = rsvg_property_bag_lookup (atts, "cy")))
+				cy = rsvg_css_parse_normalized_length (value, ctx->dpi, (gdouble)ctx->height, font_size);
+			if ((value = rsvg_property_bag_lookup (atts, "r")))
+				r = rsvg_css_parse_normalized_length (value, ctx->dpi, 
+													  rsvg_viewport_percentage((gdouble)ctx->width, (gdouble)ctx->height), 
+													  font_size);
+			if ((value = rsvg_property_bag_lookup (atts, "class")))
+				klazz = value;
+			if ((value = rsvg_property_bag_lookup (atts, "id")))
+				id = value;
+
+			rsvg_parse_style_attrs (ctx, rsvg_state_current (ctx), "circle", klazz, id, atts);
 		}
 	
 	if (r <= 0.)
-		return;
-	
-	rsvg_parse_style_attrs (ctx, rsvg_state_current (ctx), "circle", klazz, id, atts);
+		return;   
 	
 	/* approximate a circle using 4 bezier curves */
 
@@ -775,43 +929,36 @@ rsvg_start_circle (RsvgHandle *ctx, const xmlChar **atts)
 }
 
 void
-rsvg_start_ellipse (RsvgHandle *ctx, const xmlChar **atts)
+rsvg_start_ellipse (RsvgHandle *ctx, RsvgPropertyBag *atts)
 {
-	int i;
 	double cx = 0, cy = 0, rx = 0, ry = 0;
 	GString * d = NULL;
-	const char * klazz = NULL, * id = NULL;
+	const char * klazz = NULL, * id = NULL, *value;
 	char buf [G_ASCII_DTOSTR_BUF_SIZE];
 	double font_size;
 	
-	if (ctx->n_state > 0)
-		font_size = rsvg_state_current (ctx)->font_size;
-	else
-		font_size = 12.0;
+	font_size = rsvg_state_current_font_size (ctx);
 
-	if (atts != NULL)
+	if (rsvg_property_bag_size (atts))
 		{
-			for (i = 0; atts[i] != NULL; i += 2)
-				{
-					if (!strcmp ((char *)atts[i], "cx"))
-						cx = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, font_size);
-					else if (!strcmp ((char *)atts[i], "cy"))
-						cy = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, font_size);
-					else if (!strcmp ((char *)atts[i], "rx"))
-						rx = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, font_size);
-					else if (!strcmp ((char *)atts[i], "ry"))
-						ry = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, font_size);
-					else if (!strcmp ((char *)atts[i], "class"))
-						klazz = (const char *)atts[i + 1];
-					else if (!strcmp ((char *)atts[i], "id"))
-						id = (const char *)atts[i + 1];
-				}
+			if ((value = rsvg_property_bag_lookup (atts, "cx")))
+				cx = rsvg_css_parse_normalized_length (value, ctx->dpi, (gdouble)ctx->width, font_size);
+			if ((value = rsvg_property_bag_lookup (atts, "cy")))
+				cy = rsvg_css_parse_normalized_length (value, ctx->dpi, (gdouble)ctx->height, font_size);
+			if ((value = rsvg_property_bag_lookup (atts, "rx")))
+				rx = rsvg_css_parse_normalized_length (value, ctx->dpi, (gdouble)ctx->width, font_size);
+			if ((value = rsvg_property_bag_lookup (atts, "ry")))
+				ry = rsvg_css_parse_normalized_length (value, ctx->dpi, (gdouble)ctx->height, font_size);
+			if ((value = rsvg_property_bag_lookup (atts, "class")))
+				klazz = value;
+			if ((value = rsvg_property_bag_lookup (atts, "id")))
+						id = value;
+
+			rsvg_parse_style_attrs (ctx, rsvg_state_current (ctx), "ellipse", klazz, id, atts);
 		}
 	
 	if (rx <= 0. || ry <= 0.)
-		return;
-	
-	rsvg_parse_style_attrs (ctx, rsvg_state_current (ctx), "ellipse", klazz, id, atts);
+		return;   
 	
 	/* approximate an ellipse using 4 bezier curves */
 
@@ -883,12 +1030,11 @@ rsvg_start_ellipse (RsvgHandle *ctx, const xmlChar **atts)
  *         canvas 2x to get this right, only to have #1 bite us?
  */
 void
-rsvg_start_image (RsvgHandle *ctx, const xmlChar **atts)
+rsvg_start_image (RsvgHandle *ctx, RsvgPropertyBag *atts)
 {
-	int i;
 	double x = 0., y = 0., w = -1., h = -1.;
 	const char * href = NULL;
-	const char * klazz = NULL, * id = NULL;
+	const char * klazz = NULL, * id = NULL, *value;
 	
 	GdkPixbuf *img;
 	GError *err = NULL;
@@ -904,32 +1050,29 @@ rsvg_start_image (RsvgHandle *ctx, const xmlChar **atts)
 
 	state = rsvg_state_current (ctx);
 	
-	if (atts != NULL)
+	if (rsvg_property_bag_size (atts))
 		{
-			for (i = 0; atts[i] != NULL; i += 2)
-				{
-					if (!strcmp ((char *)atts[i], "x"))
-						x = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, state->font_size);
-					else if (!strcmp ((char *)atts[i], "y"))
-						y = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
-					else if (!strcmp ((char *)atts[i], "width"))
-						w = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, state->font_size);
-					else if (!strcmp ((char *)atts[i], "height"))
-						h = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
-					/* path is used by some older adobe illustrator versions */
-					else if (!strcmp ((char *)atts[i], "path") || !strcmp((char *)atts[i], "xlink:href"))
-						href = (const char *)atts[i + 1];
-					else if (!strcmp ((char *)atts[i], "class"))
-						klazz = (const char *)atts[i + 1];
-					else if (!strcmp ((char *)atts[i], "id"))
-						id = (const char *)atts[i + 1];
-				}
+			if ((value = rsvg_property_bag_lookup (atts, "x")))
+				x = rsvg_css_parse_normalized_length (value, ctx->dpi, (gdouble)ctx->width, state->font_size);
+			if ((value = rsvg_property_bag_lookup (atts, "y")))
+				y = rsvg_css_parse_normalized_length (value, ctx->dpi, (gdouble)ctx->height, state->font_size);
+			if ((value = rsvg_property_bag_lookup (atts, "width")))
+				w = rsvg_css_parse_normalized_length (value, ctx->dpi, (gdouble)ctx->width, state->font_size);
+			if ((value = rsvg_property_bag_lookup (atts, "height")))
+				h = rsvg_css_parse_normalized_length (value, ctx->dpi, (gdouble)ctx->height, state->font_size);
+			/* path is used by some older adobe illustrator versions */
+			if ((value = rsvg_property_bag_lookup (atts, "path")) || (value = rsvg_property_bag_lookup (atts, "xlink:href")))
+				href = value;
+			if ((value = rsvg_property_bag_lookup (atts, "class")))
+				klazz = value;
+			if ((value = rsvg_property_bag_lookup (atts, "id")))
+				id = value;
+
+			rsvg_parse_style_attrs (ctx, state, "image", klazz, id, atts);
 		}
 	
 	if (!href || x < 0. || y < 0. || w <= 0. || h <= 0.)
-		return;
-	
-	rsvg_parse_style_attrs (ctx, state, "image", klazz, id, atts);
+		return;   
 	
 	/* figure out if image is visible or not */
 	if (!state->visible)
@@ -996,38 +1139,34 @@ rsvg_start_image (RsvgHandle *ctx, const xmlChar **atts)
 }
 
 void 
-rsvg_start_use (RsvgHandle *ctx, const xmlChar **atts)
+rsvg_start_use (RsvgHandle *ctx, RsvgPropertyBag *atts)
 {
 	RsvgState *state = rsvg_state_current (ctx);
-	const char * klazz = NULL, *id = NULL, *xlink_href = NULL;
-	double x = 0, y = 0, width = 0, height = 0;
-	int i;
-	double affine[6];
+	const char * klazz = NULL, *id = NULL, *xlink_href = NULL, *value;
+	double x = 0, y = 0, width = 0, height = 0;	
 	gboolean got_width = FALSE, got_height = FALSE;
+	double affine[6];
 
-	if (atts != NULL)
+	if (rsvg_property_bag_size(atts))
 		{
-			for (i = 0; atts[i] != NULL; i += 2)
-				{
-					if (!strcmp ((char *)atts[i], "x"))
-						x = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->width, state->font_size);
-					else if (!strcmp ((char *)atts[i], "y"))
-						y = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
-					else if (!strcmp ((char *)atts[i], "width")) {
-						width = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
-						got_width = TRUE;
-					}
-					else if (!strcmp ((char *)atts[i], "height")) {
-						height = rsvg_css_parse_normalized_length ((char *)atts[i + 1], ctx->dpi, (gdouble)ctx->height, state->font_size);
-						got_height = TRUE;
-					}
-					else if (!strcmp ((char *)atts[i], "class"))
-						klazz = (const char *)atts[i + 1];
-					else if (!strcmp ((char *)atts[i], "id"))
-						id = (const char *)atts[i + 1];
-					else if (!strcmp ((char *)atts[i], "xlink:href"))
-						xlink_href = (const char *)atts[i + 1];
-				}
+			if ((value = rsvg_property_bag_lookup (atts, "x")))
+				x = rsvg_css_parse_normalized_length (value, ctx->dpi, (gdouble)ctx->width, state->font_size);
+			if ((value = rsvg_property_bag_lookup (atts, "y")))
+				y = rsvg_css_parse_normalized_length (value, ctx->dpi, (gdouble)ctx->height, state->font_size);
+			if ((value = rsvg_property_bag_lookup (atts, "width"))) {
+				width = rsvg_css_parse_normalized_length (value, ctx->dpi, (gdouble)ctx->height, state->font_size);
+				got_width = TRUE;
+			}
+			if ((value = rsvg_property_bag_lookup (atts, "height"))) {
+				height = rsvg_css_parse_normalized_length (value, ctx->dpi, (gdouble)ctx->height, state->font_size);
+				got_height = TRUE;
+			}
+			if ((value = rsvg_property_bag_lookup (atts, "class")))
+				klazz = value;
+			if ((value = rsvg_property_bag_lookup (atts, "id")))
+				id = value;
+			if ((value = rsvg_property_bag_lookup (atts, "xlink:href")))
+				xlink_href = value;
 		}
 	
 	/* < 0 is an error, 0 disables rendering. TODO: handle positive values correctly */
@@ -1043,23 +1182,69 @@ rsvg_start_use (RsvgHandle *ctx, const xmlChar **atts)
 					{
 					case RSVG_DEF_PATH:
 						{
-							RsvgDefsPath *path = (RsvgDefsPath*)parent;
-
-							/* combine state definitions */
-							rsvg_state_clone (state, &path->state);
-							art_affine_translate (affine, x, y);
-							art_affine_multiply (state->affine, affine, state->affine);
-
+							RsvgDefsDrawable *drawable = (RsvgDefsDrawable*)parent;
+							RsvgDefsDrawableUse * use;
+							use = g_new (RsvgDefsDrawableUse, 1);
+							use->child = drawable;
 							rsvg_parse_style_attrs (ctx, state, "use", klazz, id, atts);
-							if (state->opacity != 0xff)
-								rsvg_push_opacity_group (ctx);
-
-							/* always want to render inside of a <use/> */
-							rsvg_render_path (ctx, path->d);
-							break;
+							rsvg_state_clone (&use->super.state, state);
+							use->super.super.type = RSVG_DEF_PATH;
+							use->super.super.free = rsvg_defs_drawable_use_free;
+							use->super.draw = rsvg_defs_drawable_use_draw;
+							art_affine_translate(affine, x, y);
+							art_affine_multiply(use->super.state.affine, affine, use->super.state.affine);
+							art_affine_multiply(use->super.state.personal_affine, affine, use->super.state.personal_affine);
+							
+							if (!ctx->in_defs)
+								{
+									state = rsvg_state_current (ctx);
+									
+									rsvg_state_clone (state, &use->super.state);
+									if (ctx->n_state > 1)
+										rsvg_state_reinherit(state, &ctx->state[ctx->n_state - 2]);
+									
+									if (state->opacity != 0xff || state->filter)
+										rsvg_push_discrete_layer (ctx);
+									
+									
+									/* push the state stack */
+									if (ctx->n_state == ctx->n_state_max)
+										ctx->state = g_renew (RsvgState, ctx->state, 
+															  ctx->n_state_max <<= 1);
+									if (ctx->n_state)
+										rsvg_state_inherit (&ctx->state[ctx->n_state],
+															&ctx->state[ctx->n_state - 1]);
+									else
+										rsvg_state_init (ctx->state);
+									ctx->n_state++;
+									
+									rsvg_defs_drawable_draw (use->child, ctx);
+									
+									/* pop the state stack */
+									ctx->n_state--;
+									rsvg_state_finalize (&ctx->state[ctx->n_state]);
+									
+									if (state->opacity != 0xff || state->filter)
+										rsvg_pop_discrete_layer (ctx);
+									break;
+								}							
+							else
+								{
+									if (id != NULL)
+										rsvg_defs_set (ctx->defs, id, &use->super.super);
+									
+									use->super.parent = (RsvgDefsDrawable *)ctx->current_defs_group;
+									if (use->super.parent != NULL)
+										rsvg_defs_drawable_group_pack((RsvgDefsDrawableGroup *)use->super.parent, 
+																	  &use->super);
+									
+									
+									break;
+								}
 						}
 					default:
-						g_warning ("Unhandled defs entry/type %s %d\n", id, parent->type);
+						g_warning ("Unhandled defs entry/type %s %d\n", id, 
+								   parent->type);
 						return;
 					}
 		}
